from matplotlib import pyplot as plt
import numpy as np
import math


def celltype_visualization_ST(pred_p, pos_all, col_num,plot=True):
    """
    pred_p: dataframe, spot * cell_type_num
    pos_all: dataframe, spot * 2
    """
    x_num = pos_all.iloc[:, 0].max() - pos_all.iloc[:, 0].min() + 1
    y_num = pos_all.iloc[:, 1].max() - pos_all.iloc[:, 1].min() + 1
    cell_type_num = pred_p.shape[1]
    cell_type_pred_map = np.zeros([x_num, y_num, cell_type_num])
    for i in range(len(pos_all)):
        row_i = int(pos_all.iloc[i, 0])
        col_i = int(pos_all.iloc[i, 1])
        cell_type_pred_map[row_i, col_i, :] = pred_p.iloc[i, :]

    row_num = math.ceil(cell_type_num / col_num)
    if row_num == 1:
        row_num = 2
    fig, ax = plt.subplots(row_num, col_num, figsize=(8, 8))

    for i in range(row_num):
        for j in range(col_num):
            ax[i, j].axis('off')
            ax[i, j].set_aspect('equal')

    for i in range(pred_p.shape[1]):
        row_i = i // col_num
        col_i = i - row_i * col_num
        im = ax[row_i, col_i].scatter(pos_all.iloc[:, 0], pos_all.iloc[:, 1], s=20, c=pred_p.iloc[:, i], marker='.', cmap='magma')
        plt.colorbar(im, ax=ax[row_i, col_i], cmap='magma')
        ax[row_i, col_i].set_title(pred_p.columns[i], fontdict={'fontsize': 20, 'fontname': 'Arial'})
        ax[row_i, col_i].set_aspect(np.sqrt(3))

    if plot:
        plt.show()


def celltype_visualization_visium(pred_p, pos_all, col_num):
    """
    pred_p: dataframe, spot * cell_type_num
    pos_all: dataframe, spot * 2
    """

    pos_all = pos_all.iloc[:, [1, 0]]
    pos_all.iloc[:, 1] = pos_all.iloc[:, 1].max() - pos_all.iloc[:, 1]

    cell_type_num = pred_p.shape[1]
    row_num = math.ceil(cell_type_num / col_num)
    if row_num == 1:
        row_num = 2
    fig, ax = plt.subplots(row_num, col_num, figsize=(30, 12))

    for i in range(row_num):
        for j in range(col_num):
            ax[i, j].axis('off')
            ax[i, j].set_aspect('equal')

    for i in range(pred_p.shape[1]):
        row_i = i // col_num
        col_i = i - row_i * col_num
        im = ax[row_i, col_i].scatter(pos_all.iloc[:, 0], pos_all.iloc[:, 1], s=20, c=pred_p.iloc[:, i], marker='.', cmap='magma')
        plt.colorbar(im, ax=ax[row_i, col_i], cmap='magma')
        ax[row_i, col_i].set_title(f'X{pred_p.columns[i]}', fontdict={'fontsize': 40, 'fontname': 'Arial'})
        ax[row_i, col_i].set_aspect(np.sqrt(3))
    plt.show()


def visualize_lineplot(cell_type_num, index, rare_cell_type_num):
    fig, ax1 = plt.subplots(figsize=(3, 2))
    ax1.plot(cell_type_num, index, color='#416FB9')
    ax1.set_xlabel('K', fontname='Arial', fontsize=9)
    ax1.set_ylabel('Pereplexity', color='#416FB9', fontsize=9)
    ax1.tick_params('y', colors='#416FB9')
    ax1.tick_params(axis='both', which='both', labelsize=6)

    ax2 = ax1.twinx()
    ax2.plot(cell_type_num, rare_cell_type_num, color='#948434')
    ax2.set_ylabel('Rare cell type number', color='#948434', fontsize=9)
    ax2.tick_params('y', colors='#948434')
    ax2.tick_params(axis='both', which='both', labelsize=6)

    plt.show()


import collections
import tables
import scipy.sparse as sp_sparse


def get_data_from_h5(filename):
    CountMatrix = collections.namedtuple('CountMatrix', ['feature_ref', 'barcodes', 'matrix'])
    with tables.open_file(filename, 'r') as f:
        mat_group = f.get_node(f.root, 'matrix')
        barcodes = f.get_node(mat_group, 'barcodes').read()
        data = getattr(mat_group, 'data').read()
        indices = getattr(mat_group, 'indices').read()
        indptr = getattr(mat_group, 'indptr').read()
        shape = getattr(mat_group, 'shape').read()
        matrix = sp_sparse.csc_matrix((data, indices, indptr), shape=shape)

        feature_ref = {}
        feature_group = f.get_node(mat_group, 'features')
        feature_ids = getattr(feature_group, 'id').read()
        feature_names = getattr(feature_group, 'name').read()
        feature_types = getattr(feature_group, 'feature_type').read()
        feature_ref['id'] = feature_ids
        feature_ref['name'] = feature_names
        feature_ref['feature_type'] = feature_types
        tag_keys = getattr(feature_group, '_all_tag_keys').read()
        for key in tag_keys:
            feature_ref[key.decode('utf-8')] = getattr(feature_group, key.decode('utf-8')).read()

        return CountMatrix(feature_ref, barcodes, matrix)


import pandas as pd
from utils.data_pp import spatial_pp
import numpy as np
from utils.deconvolution import deconvolution
from utils.calculate_cell_type_expression import calc_cell_type_expr
from utils.cell_type_matching import cell_type_matching
import seaborn as sns
import matplotlib.pyplot as plt
from utils.cell_type_visualization import celltype_visualization_ST

data_dir = 'example_data/simulated_data/raw_data/'
save_dir = 'example_data/simulated_data/pp_data/'
expr_path = data_dir + '/count.csv'
pos_path = data_dir + '/pos.csv'

df_expr = pd.read_csv(expr_path)
df_pos = pd.read_csv(pos_path)
df_pos['x'] = df_pos['x'] - df_pos['x'].min()
df_pos['y'] = df_pos['y'] - df_pos['y'].min()
barcodes = np.array(range(len(df_expr)))

df_pp_data = spatial_pp(df_expr, df_pos, barcodes, save_dir)

df_deconvolution_results, df_beta = deconvolution(df_pp_data, cell_type_num=4, spatial_mode='square', save_dir_name='simulated_data', margin=0.05, device_id=0, num_epoch=5)

# Here we use the deconvolution results of epoch=500 for further exploration.
results_dir = 'results_save/simulated_data_example1'
df_deconvolution_results = pd.read_csv(results_dir + '/prediction_save/pred.csv')
df_beta = pd.read_csv(results_dir + '/prediction_save/beta.csv')

sc_data_path = data_dir + 'sc_data.csv'
df_sc = pd.read_csv(sc_data_path)
df_gt_expr = calc_cell_type_expr(df_sc, save_dir)
df_matching_results, df_corr = cell_type_matching(df_gt_expr, df_beta)

plt.figure(figsize=(5, 4))
sns.heatmap(df_corr, annot=True, cmap='YlGnBu', linewidths=0.5)
plt.title('Cell type matching')
plt.show()




import os
import torch
import torch.nn.functional as F
import numpy as np
from .wae import WAE
import sys
import torch.nn as nn
from torch.nn.functional import cosine_similarity
import collections
import math
from torch.cuda.amp import autocast, GradScaler
import time
import torch.optim.lr_scheduler as lr_scheduler
import random


class WTM:
    def __init__(self, bow_dim=10000, n_topic=20, device=None, dropout=0.0, alpha=0.2):
        self.bow_dim = bow_dim
        self.n_topic = n_topic
        self.wae = WAE(encode_dims=[bow_dim, 50, 30, n_topic], decode_dims=[n_topic, 50, bow_dim], dropout=dropout, device=device)
        self.wae.initialize()
        self.device = device
        self.alpha = np.full([n_topic, ], alpha)

        if device != None:
            self.wae = self.wae.to(device)

    def train(self, train_dataloader, num_epochs=100, save_dir=None, margin=0.05):
        optimizer = torch.optim.AdamW(self.wae.parameters(), lr=1)
        scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda epoch: 0.01 if epoch < 200 else 0.001)

        start_epoch = 0
        for epoch in range(start_epoch, num_epochs):
            train_loss = 0
            train_rec_loss = 0
            train_cos_loss = 0
            train_kl_loss = 0
            train_mmd_loss = 0
            train_contras_loss = 0
            train_num = 0

            self.wae.train()


            for iter, data in enumerate(train_dataloader):
                optimizer.zero_grad()
                bows = data[0].to(self.device)
                positive_sample = data[1].to(self.device)
                negative_sample = data[2].to(self.device)
                pos_num = data[3]
                neg_num = data[4]


                with autocast(enabled=False):
                    bows_recon, theta_q, weight = self.wae(bows)
                    _, theta_q_pos, _ = self.wae(positive_sample)
                    _, theta_q_neg, _ = self.wae(negative_sample)

                    theta_prior = self.wae.sample(alpha=self.alpha, batch_size=len(bows)).to(self.device)

                    bows_recon = F.softmax(bows_recon, dim=1)
                    cos_loss = (1 - cosine_similarity(bows_recon, bows, axis=1)).sum() / bows.shape[0]
                    kl_loss_calc = nn.KLDivLoss(reduction='batchmean', log_target=True)
                    kl_loss = kl_loss_calc(torch.log(bows_recon), torch.log(bows + 1e-6))
                    rec_loss = (cos_loss + kl_loss) * 0.5
                    mmd = self.wae.mmd_loss(theta_q.squeeze(), theta_prior, device=self.device, t=0.1) * 10.0
                    contras_loss = self.wae.contrastive_loss_triplet(theta_q, theta_q_pos, theta_q_neg, pos_num.to(self.device),
                                                                     neg_num.to(self.device), margin=margin, device=self.device)

                    loss = rec_loss + mmd + contras_loss


                loss.backward()
                optimizer.step()

                train_loss += loss.item() * bows.shape[0]
                train_rec_loss += rec_loss.item() * bows.shape[0]
                train_cos_loss += cos_loss.item() * bows.shape[0]
                train_kl_loss += kl_loss.item() * bows.shape[0]
                train_mmd_loss += mmd.item() * bows.shape[0]
                train_contras_loss += contras_loss.item() * bows.shape[0]
                train_num += bows.size(0)

            if epoch == num_epochs - 1:
                last_model = self.wae.state_dict()
            scheduler.step()

            print('Epoch {}/{} completed'.format(epoch + 1, num_epochs))
        if save_dir != None:
            torch.save(last_model, os.path.join(save_dir, 'last.pkl'))


    def inference_by_bow(self, spot_fea):
        with torch.no_grad():
            self.wae.eval()
            with autocast(enabled=False):
                theta = F.softmax(self.wae.encode(spot_fea.to(self.device)), dim=-1)
                bows_recon, _, _ = self.wae(spot_fea.to(self.device))
            return theta.detach().cpu().numpy(), bows_recon.detach().cpu().numpy()

    def theta_filter(self, result, thres=0.01):
        result[result < thres] = 0
        row_sum = result.sum(axis=1)
        result = result / row_sum[:, np.newaxis]

        return result

    def get_embed(self, dataloader):
        self.wae.eval()
        embed_lst = []
        recon_lst = []
        for data_batch in dataloader:
            embed, recon = self.inference_by_bow(data_batch)
            embed_lst.append(embed)
            if len(recon.shape) < 2:
                recon = recon[np.newaxis, :]
            recon_lst.append(recon)
        embed_lst = np.concatenate(embed_lst, axis=0)
        recon_lst = np.concatenate(recon_lst, axis=0)
        # embed_lst = self.theta_filter(embed_lst, thres=0.05)
        return embed_lst, recon_lst


    def get_topic_top_words(self, gene_names_list, top_k=5):
        self.wae.eval()
        decoder_w = self.wae.state_dict()['decoder.0.weight'].T.detach().cpu()
        beta = F.softmax(decoder_w, dim=1)
        params_new = -np.array(beta)
        top_gene_ids = np.argsort(params_new)[:, :top_k]
        top_gene_names = np.empty_like(top_gene_ids, dtype='U100')

        for i in range(top_gene_ids.shape[0]):
            for j in range(top_gene_ids.shape[1]):
                top_gene_names[i][j] = gene_names_list[top_gene_ids[i][j]]

        return top_gene_ids, top_gene_names, beta


    def calc_rare_cell_type_num(self, embeds, thres=0.05):
        proportion_mean = np.mean(embeds, axis=0)
        rare_num = np.sum(proportion_mean < thres)
        rare_cell_type = np.where(proportion_mean < thres)

        return rare_num, rare_cell_type


    def calc_perplexity(self, test_data, recon_data):  # test_data pp
        logsoftmax = torch.log_softmax(torch.from_numpy(recon_data), dim=1)
        J = -1.0 * torch.sum(torch.from_numpy(np.array(test_data)) * logsoftmax) / len(test_data)

        perplexity = math.exp(J)
        return perplexity





import pandas as pd
from utils.data_pp import spatial_pp
from utils.deconvolution import deconvolution_multi_ctn
from utils.get_data_from_h5 import get_data_from_h5
from utils.cell_type_visualization import visualize_lineplot, celltype_visualization_visium

data_dir = 'example_data/CRLM_visium/raw_data/'
save_dir = 'example_data/simulated_data/pp_data/'
expr_path = data_dir + '/filtered_feature_bc_matrix.h5'
pos_path = data_dir + '/tissue_positions_list.csv'

filtered_matrix_h5 = get_data_from_h5(expr_path)
df_expr = pd.DataFrame(filtered_matrix_h5.matrix.A, columns=filtered_matrix_h5.barcodes.astype(str), index=filtered_matrix_h5.feature_ref['name'].astype(str))
df_expr = df_expr.T
df_pos = pd.read_csv(pos_path, index_col=0, header=None)
df_pos = df_pos.loc[df_pos.iloc[:, 0] == 1].iloc[:, 1:3]
df_pos.rename(columns={df_pos.columns[0]: 'x'}, inplace=True)
df_pos.rename(columns={df_pos.columns[1]: 'y'}, inplace=True)
df_pos['x'] = df_pos['x'] - df_pos['x'].min()
df_pos['y'] = df_pos['y'] - df_pos['y'].min()
df_expr = df_expr.reindex(df_pos.index)

df_expr.reset_index(drop=True, inplace=True)
df_pos.reset_index(drop=True, inplace=True)
barcodes = df_pos.index.to_numpy()

df_pp_data = spatial_pp(df_expr, df_pos, barcodes, save_dir)
df_deconvolution_data = deconvolution_multi_ctn(df_pp_data, cell_type_num=[2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15], spatial_mode='hexagon', save_dir_name='CRLM_visium', margin=0.05, device_id=0, num_epoch=2)

# Here we use the deconvolution results of epoch=500 for further exploration.
results_dir = 'results_save/CRLM_visium_example2'
df_deconvolution_data = pd.read_csv(results_dir + '/CRLM_visium.csv')

# Plot cell type number K versus perplexity and rare cell type number
cell_type_num = df_deconvolution_data.iloc[:, 0]
ppl = df_deconvolution_data.iloc[:, 2]
rare_cell_type_num = df_deconvolution_data.iloc[:, 1]
visualize_lineplot(cell_type_num, ppl, rare_cell_type_num)

# Choose a suitable cell_type_num for further visualization (The optimal cell type number K is determined at the point where reaches the lowest perplexity while minimizing the rare cell type number ideally. It is reconmmended to choose the cell type number with a low perplexity but not very high rare cell type number (less than 3). Moreover, the user can choose the optimal K with the help their biology knowledge.)
cell_type_num = 11
df_pred_ = pd.read_csv(results_dir + f'/prediction_save/cell_type_num_{cell_type_num}/pred.csv')
df_pos_ = df_pp_data.iloc[:, 1:3]
celltype_visualization_visium(df_pred_, df_pos_, col_num=6)

